@@ -5,7 +5,7 @@ use pyo3::exceptions::PyIndexError;
 use pyo3::pyclass::CompareOp;
 use pyo3::types::{PyDict, PyIterator, PyTuple, PyType};
 use pyo3::{exceptions::PyKeyError, types::PyMapping};
-use pyo3::{prelude::*, AsPyPointer};
+use pyo3::{prelude::*, AsPyPointer, PyTypeInfo};
 use rpds::{HashTrieMap, HashTrieMapSync, HashTrieSet, HashTrieSetSync, List, ListSync};
 
 #[derive(Clone, Debug)]
@@ -39,7 +39,7 @@ impl IntoPy<PyObject> for Key {
     }
 }
 
-impl AsPyPointer for Key {
+unsafe impl AsPyPointer for Key {
     fn as_ptr(&self) -> *mut pyo3::ffi::PyObject {
         self.inner.as_ptr()
     }
@@ -133,7 +133,7 @@ impl HashTrieMapPy {
         let contents = self.inner.into_iter().map(|(k, v)| {
             format!(
                 "{}: {}",
-                k.into_py(py),
+                k.clone().into_py(py),
                 v.call_method0(py, "__repr__")
                     .and_then(|r| r.extract(py))
                     .unwrap_or("<repr error>".to_owned())
@@ -167,6 +167,16 @@ impl HashTrieMapPy {
         }
     }
 
+    fn __reduce__(slf: PyRef<Self>) -> (&PyType, (Vec<(Key, PyObject)>,)) {
+        (
+            HashTrieMapPy::type_object(slf.py()),
+            (slf.inner
+                .iter()
+                .map(|(k, v)| (k.clone(), v.clone()))
+                .collect(),),
+        )
+    }
+
     #[classmethod]
     fn convert(_cls: &PyType, value: &PyAny, py: Python) -> PyResult<PyObject> {
         if value.is_instance_of::<HashTrieMapPy>() {
@@ -188,8 +198,8 @@ impl HashTrieMapPy {
         self.inner.values().collect::<Vec<&PyObject>>()
     }
 
-    fn items(&self) -> Vec<(&Key, &PyObject)> {
-        self.inner.iter().collect::<Vec<(&Key, &PyObject)>>()
+    fn items(&self) -> Vec<(Key, &PyObject)> {
+        self.inner.iter().map(|(k, v)| (k.clone(), v)).collect()
     }
 
     fn discard(&self, key: Key) -> PyResult<HashTrieMapPy> {
@@ -318,7 +328,8 @@ impl HashTrieSetPy {
 
     fn __repr__(&self, py: Python) -> String {
         let contents = self.inner.into_iter().map(|k| {
-            k.into_py(py)
+            k.clone()
+                .into_py(py)
                 .call_method0(py, "__repr__")
                 .and_then(|r| r.extract(py))
                 .unwrap_or("<repr failed>".to_owned())
@@ -345,6 +356,13 @@ impl HashTrieSetPy {
         }
     }
 
+    fn __reduce__(slf: PyRef<Self>) -> (&PyType, (Vec<Key>,)) {
+        (
+            HashTrieSetPy::type_object(slf.py()),
+            (slf.inner.iter().cloned().collect(),),
+        )
+    }
+
     fn insert(&self, value: Key) -> HashTrieSetPy {
         HashTrieSetPy {
             inner: self.inner.insert(value),
@@ -545,6 +563,13 @@ impl ListPy {
         }
     }
 
+    fn __reduce__(slf: PyRef<Self>) -> (&PyType, (Vec<PyObject>,)) {
+        (
+            ListPy::type_object(slf.py()),
+            (slf.inner.iter().cloned().collect(),),
+        )
+    }
+
     #[getter]
     fn first(&self) -> PyResult<&PyObject> {
         self.inner
@@ -552,17 +577,25 @@ impl ListPy {
             .ok_or_else(|| PyIndexError::new_err("empty list has no first element"))
     }
 
+    #[getter]
+    fn rest(&self) -> ListPy {
+        let mut inner = self.inner.clone();
+        inner.drop_first_mut();
+        ListPy { inner }
+    }
+
     fn push_front(&self, other: PyObject) -> ListPy {
         ListPy {
             inner: self.inner.push_front(other),
         }
     }
 
-    #[getter]
-    fn rest(&self) -> ListPy {
-        let mut inner = self.inner.clone();
-        inner.drop_first_mut();
-        ListPy { inner }
+    fn drop_first(&self) -> PyResult<ListPy> {
+        if let Some(inner) = self.inner.drop_first() {
+            Ok(ListPy { inner })
+        } else {
+            Err(PyIndexError::new_err("empty list has no first element"))
+        }
     }
 }
 
